/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Debug.h
* 
* AUTHOR: 
*
* Stephan R.A. Deibel
* 
* CREATION DATE: 
*
* 8/1/93
* 
* VERSION: 
*
* 6/13/94
* 
* DESCRIPTION: 
* 
* Basic debug utilities that replace some of the standard C "assert"
* implementations with somewhat safer macros.
*
* For usage instructions read the embedded comments below.
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 14May93  Stephan   Created
* 14Jun94  Stephan   Cosmetic / commenting
*
*/

#ifndef CU_DEBUG_H_ALREADY_INCLUDED
#define CU_DEBUG_H_ALREADY_INCLUDED

#include "CU_Basics.h"

#ifdef macintosh
#include <Types.h>
#include <SegLoad.h>
#endif /* macintosh */

#ifdef __cplusplus
extern "C" {
#endif


/**************************************************************************/
/* TYPES */

typedef struct CUt_DBMallocInfo {
  void* ptr;
  const char* file_name;
  unsigned line_num;
  unsigned long size;
  struct CUt_DBMallocInfo* next;
} CUt_DBMallocInfo;


/**************************************************************************/
/* PROTOTYPES */

EXPORT_PREFIX void * _DB_Malloc(const char *, unsigned int, unsigned long);
EXPORT_PREFIX void
_DB_Free(const char *, unsigned int, void*);
EXPORT_PREFIX void*
_DB_Realloc(const char *, unsigned long, void*, unsigned long);
EXPORT_PREFIX void*
_DB_Calloc(const char *, unsigned long, unsigned long, unsigned long);
EXPORT_PREFIX void*
_DB_MallocSet(const char *, unsigned long, unsigned long, unsigned char);
EXPORT_PREFIX tUInt2
CU_DumpDBMallocInfo(const char* file_name);


/**************************************************************************/
/* MACROS */


/*************************************************************************** *
* MACRO:
*
* UNUSED()
*
* DESCRIPTION:
*
* Assert a parameter as intentionally unused.  Avoids compiler warnings
* and makes it clearer that the parameter is needed in the code due
* to type requirements but is not needed in an implementation.
*
* PARAMETERS:
*
* Any type parameter
*
*/

#ifndef UNUSED
#define UNUSED(p) { if (p) p=p; }
#endif

/*************************************************************************** *
* MACRO:
*
* ASSERT()
*
* DESCRIPTION:
*
* DEBUG mode assertion macro to replace standard C "assert()"
*
* PARAMETERS:
*
* One boolean expression of any type.
*
* NOTES:
*
* This macro should be used to assert assumed conditions as TRUE and to
* check for programming errors in your code.  It is good practice to
* include a set of assertions in every routine that you write and to
* leave them in the code throughout the life of that code, even after
* you think you have debugged it.  Future changes or improper calls
* to your code may reveal additional bugs.
*
* When DEBUG is defined, the macro will enter the resident debugger when
* an assertion has failed.  When DEBUG is not defined, the ASSERT
* macro ceases to exist so that your shipping version of your application
* is not slowed down by debug code.
*
* The proper form for using ASSERT is to place the statement on a single
* line by itself.  Extreme caution should be exersized to avoid using
* ASSERT(f) where f() has some side effect on data structures.  If this
* occurs, then undefining DEBUG may introduce bugs into your program.
*
* The DEBUG version of your application MUST be a STRICT NON-INTERFERING
* SUPERSET of your shipping version!
*
* Examples:
*
* ASSERT(theHandle <> nil);   / * Check parameter passed in for validity * /
* ASSERT(i > 0);              / * Check intermediate result for validity * /
* ASSERT(newNode = malloc(sizeof(tNode)));        / * DO NOT DO THIS!!! * /
*
* The last example above is bad because:
*
* 1) it violates the "strict non-interfering superset" approach to using
*    ASSERT (the contents of the ASSERT has side effects that affect the
*    overall state of your program).  This is particularly a problem
*    because the statement will cease to exist in your shipping application 
*    so that the "malloc" will never be called.
* 2) ASSERT is being used to check for acceptible failure.  In your
*    debug application, your program will essentially crash when out
*    of memory.  In your shipping application, your program will not
*    be checking for out-of-memory.  Instead, you should use a regular
*    "if" and appropriate out-of-memory exit condition for your routine.
*
*/

#ifdef DEBUG
#ifdef ASSERT
#undef ASSERT
#endif
	void EXPORT_PREFIX _DB_Assert(const char *, unsigned int);
	#define ASSERT(f) { if (!(f)) _DB_Assert(__FILE__,__LINE__); }
#else
	#define ASSERT(f) 
#endif


/*************************************************************************** *
* MACRO:
*
* BARF  
*
* DESCRIPTION:
*
* DEBUG mode special case of ASSERT macro for a guaranteed failure
* condition.  
*
* NOTES:
* 
* Exactly the same as ASSERT(0); 
* 
*/

#ifdef DEBUG
	#define BARF _DB_Assert(__FILE__,__LINE__)
#else
	#define BARF 
#endif


/*************************************************************************** *
* MACRO:
*
* UNIMPLEMENTED
*
* DESCRIPTION:
*
* DEBUG mode macro for marking unimplemented calls.
*
* NOTES:
*
* This macro should be used in functions (or sub-parts of functions) that have
* been defined but not yet implemented.  When DEBUG is defined, the macro will 
* enter the resident debugger and indicate that an unimplemented routine has 
* been encountered.  When DEBUG is not define, the UNIMPLEMENTED macro ceases 
* to exist.
*
* The proper form for using UNIMPLEMENTED is to place the statement on a single
* line by itself:
*
* Example:
*
* UNIMPLEMENTED;
*
*/

#ifdef DEBUG
	void EXPORT_PREFIX _DB_Unimplemented(const char *, unsigned int);
	#define UNIMPLEMENTED _DB_Unimplemented(__FILE__,__LINE__)
#else
	#define UNIMPLEMENTED 
#endif



/*************************************************************************** *
* MACRO:
*
* UNFINISHED
*
* DESCRIPTION:
*
* DEBUG mode macro for marking unfinished calls without halting the
* program as UNIMPLEMENTED will.
*
* NOTES:
*
* This macro should be used in functions (or sub-parts of functions) that
* have een defined but not yet implemented.  When DEBUG is defined, the
* macro will put an error message to stderr but will not exit the
* running of the program. The UNFINISHED macro can be used to mark
* non-essential routines without being forced to implement them
* immediately. 
*
* The use of the macro is simlar to that of UNIMPLEMENTED, that is to
* place it in a line by itself:
*
* Example:
*
* UNFINISHED;
*
*/

#ifdef DEBUG
	void EXPORT_PREFIX _DB_Unfinished(const char *, unsigned int);
	#define UNFINISHED _DB_Unfinished(__FILE__,__LINE__)
#else
	#define UNFINISHED 
#endif


/*************************************************************************** *
* MACRO:
*
* UNTESTED
*
* DESCRIPTION:
*
* DEBUG mode macro for marking untested code.  This will simply issue
* a warning and continue execution.
*
* NOTES:
*
* This macro should be used in functions (or sub-parts of functions) that 
* have been implemented but not yet tested.  When DEBUG is defined, the
* macro will put a warning message to stderr but will not exit the
* running of the program.
*
* The use of the macro is simlar to that of UNIMPLEMENTED, that is to
* place it in a line by itself:
*
* Example:
*
* UNTESTED;
*
*/

#ifdef DEBUG
	void EXPORT_PREFIX _DB_Untested(const char *, unsigned int);
	#define UNTESTED _DB_Untested(__FILE__,__LINE__)
#else
	#define UNTESTED 
#endif


/*************************************************************************** *
* MACRO:
*
* DEBUGMSG
*
* DESCRIPTION:
*
* DEBUG mode macro for printing a message to stderr.
*
* NOTES:
*
* When DEBUG is not defined, this macro will produce no code.
*
* Example:
*
* DEBUGMSG;
*
*/

#ifdef DEBUG
	void EXPORT_PREFIX _DB_DebugMsg(const char *, unsigned int, const char *msg);
	#define DEBUGMSG(msg) _DB_DebugMsg(__FILE__,__LINE__,msg)
#else
	#define DEBUGMSG(msg) 
#endif


/*************************************************************************** *
* MACRO:
*
* GOODMEM
*
* DESCRIPTION:
*
* DEBUG mode macro for checking that a pointer allocated in the free
* store is pointing to good memory.  Will print a debug message and
* halt when the pointer is bad.
*
* NOTES:
*
* When DEBUG is not defined, this macro will produce no code.
*
* Currently, this only checks if the pointer is NULL; no more
* sophisticated check is implemented.
*
* Example:
*
* GOODMEM(my_ptr);
*
*/

#ifdef DEBUG
	void EXPORT_PREFIX _DB_GoodMem(const char *, unsigned int, const void *ptr);
	#define GOODMEM(ptr) _DB_GoodMem(__FILE__,__LINE__, (const void *) ptr)
#else
	#define GOODMEM(ptr) 
#endif


/*************************************************************************** *
* MACRO:
*
* CU_Malloc/CU_Free/CU_Realloc/CU_Calloc/CU_MallocSet
*
* DESCRIPTION:
*
* These macros convert calls to the CU set of memory functions to equivalent
* _DB functions with argument to capture point of call info.  They are only
* defined when CUk_DebugMalloc is defined; this allows parts of a project 
* which should free all the memory they allocate to use the debugging memory 
* functions
*
* NOTES:
*
* The current implementation simply uses a linked list, and _DB_Realloc and
* _DB_Free do a linear search through the list.  We may want to use some sort
* of hashing in the future, but we can't use the CU_Hash utils because they
* may be using the debugging memory functions (causing infinite recursion).
*
*/

#ifdef DEBUG
 #ifdef CUk_DebugMalloc
        #define CU_Malloc(size) _DB_Malloc(__FILE__,__LINE__, (size))
        #define CU_Free(ptr) _DB_Free(__FILE__,__LINE__, (ptr))
        #define CU_Realloc(ptr, size) _DB_Realloc(__FILE__,__LINE__, (ptr), (size))
        #define CU_Calloc(num, size) _DB_Calloc(__FILE__,__LINE__, (num), (size))
        #define CU_MallocSet(size, fill) _DB_MallocSet(__FILE__,__LINE__, (size), (fill))
 #endif
#endif

#ifdef __cplusplus
} /* extern "C" */
#endif 

#endif /* CU_DEBUG_H_ALREADY_INCLUDED */
