/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Hash.h
* 
* AUTHOR: 
*
* YongJoon Lee
* John Ehresman
* Based on work of John K. Ousterhaut
* 
* CREATION DATE: 
*
* 95.9.22
*
* DESCRIPTION: 
*
* Public header file for hash table implementation.
* 
* NOTES: 
*
* Based on internal tcl routines v. 7.4b2.  The original copyright is 
* as follows:
*
*   Copyright (c) 1987-1994 The Regents of the University of California.
*   Copyright (c) 1994 Sun Microsystems, Inc.
*
*   See the file "tcl_license.terms" for information on usage and 
*   redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*
* Names have been modified to conform to the Arachne standard.
* 
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* $Log: CU_Hash.h,v $
* Revision 1.3  1998/12/01 23:18:49  bsloan
* Port to MetroWerks CodeWarrior Pro4 for Windows
*
* Revision 1.2  1998/07/15 05:20:15  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.14  1997/02/06 02:02:07  jehresma
* Fixed comment on CU_GetHashEntryCount macro
*
* Revision 1.13  1997/02/06 02:01:01  jehresma
* Added CU_GetHashEntryCount macro
*
* Revision 1.12  1996/12/09 19:22:56  sdeibel
* Misc clean up
*
* 
*/

#ifndef CU_HASH_H_ALREADY_INCLUDED
#define CU_HASH_H_ALREADY_INCLUDED


#include <stdio.h>
#if !defined(__MWERKS__) && !defined(WIN32)
	#include <memory.h>
#endif // !(__MWERKS__ && WIN32)
#include <string.h>

#ifdef dos
#include "CUBasics.h"
#include "CUFail.h"
#include "CUDebug.h"
#else
#include "CU_Basics.h"
#include "CU_Failure.h"
#include "CU_Debug.h"
#endif /* dos */

#ifdef __cplusplus
extern "C" {
#endif


/**************************************************************************/
/* CONSTANTS */

/* Acceptable key types for hash tables: */
#define CUk_STRING_KEYS		0
#define CUk_ARRAY_KEYS          4

/* Initial size for small hash table */
#define CUk_SMALL_HASH_TABLE    4


/**************************************************************************/
/* TYPES */


typedef void * CUt_ClientData;

/* Forward declaration of CUt_HashTable.  Needed by some C++ compilers
   to prevent errors when the forward reference to CUt_HashTable is
   encountered in the CUt_HashEntry structure. */
#ifdef __cplusplus
struct CUt_HashTable;
#endif

/* Structure definition for an entry in a hash table.  No-one outside
   Tcl should access any of these fields directly;  use the macros
   defined below. */
typedef struct CUt_HashEntry {
  struct CUt_HashEntry *nextPtr;	/* Pointer to next entry in this
					 * hash bucket, or NULL for end of
					 * chain. */
  struct CUt_HashTable *tablePtr;	/* Pointer to table containing entry. */
  struct CUt_HashEntry **bucketPtr;	/* Pointer to bucket that points to
					 * first entry in this entry's chain:
					 * used for deleting the entry. */
  CUt_ClientData clientData;		/* Application stores something here
					 * with CU_SetHashValue. */
  union {				/* Key has one of these forms: */
    tInt1 bytes[4];			/* Multiple integer words for key.
					 * The actual size will be as large
					 * as necessary for this table's
					 * keys. */
    char string[4];			/* String for key.  The actual size
					 * will be as large as needed to hold
					 * the key. */
  } key;				/* MUST BE LAST FIELD IN RECORD!! */
} CUt_HashEntry;

/* Structure definition for a hash table.  Must be in tcl.h so clients
   can allocate space for these structures, but clients should never
   access any fields in this structure. */
typedef struct CUt_HashTable {
  CUt_HashEntry **buckets;		/* Pointer to bucket array.  Each
					 * element points to first entry in
					 * bucket's hash chain, or NULL. */
  CUt_HashEntry *staticBuckets[CUk_SMALL_HASH_TABLE];
					/* Bucket array used for small tables
					 * (to avoid mallocs and frees). */
  tUInt4 numBuckets;			/* Total number of buckets allocated
					 * at **bucketPtr. */
  tUInt4 numEntries;			/* Total number of entries present
					 * in table. */
  tUInt4 rebuildSize;			/* Enlarge table when numEntries gets
					 * to be this large. */
  tInt4 downShift;			/* Shift count used in hashing
					 * function.  Designed to use high-
					 * order bits of randomized keys. */
  tInt4 mask;				/* Mask value used in hashing
					 * function. */
  tUInt4 keyType;			/* Type of keys used in this table. 
					 * It's either CUk_STRING_KEYS,
					 * CUk_ONE_WORD_KEYS, or an integer
					 * giving the number of ints in a
					 */
  CUt_HashEntry *(*findProc) (const struct CUt_HashTable *tablePtr, 
                              const void *key);
  CUt_HashEntry *(*createProc) (struct CUt_HashTable *tablePtr,
                                const void *key, tInt4 *newPtr);
} CUt_HashTable;

/* Structure definition for information used to keep track of searches
   through hash tables: */
typedef struct CUt_HashSearch {
    CUt_HashTable *tablePtr;		/* Table being searched. */
    tUInt4 nextIndex;			/* Index of next bucket to be
					 * enumerated after present one. */
    CUt_HashEntry *nextEntryPtr;	/* Next entry to be enumerated in the
					 * the current bucket. */
} CUt_HashSearch;


/**************************************************************************/
/* ACCESS MACROS */

/* PROTO:
* void *CU_GetHashValue(CUt_HashEntry *h);
*/
#define CU_GetHashValue(h) ((h)->clientData)

/* PROTO:
* void CU_SetHashValue(CUt_HashEntry *h, void *value);
*/
#define CU_SetHashValue(h, value) ((h)->clientData = (CUt_ClientData) (value))


/* PROTO:
* char *CU_GetHashKey(CUt_HashTable *tablePtr, CUt_HashEntry *h);
*/ 
#define CU_GetHashKey(tablePtr, h) \
    ((char *) (((tablePtr)->keyType == CUk_STRING_KEYS) ? (h)->key.string \
               : ((h)->key.bytes)))

/* PROTO:
* tUInt4 CU_GetHashEntryCount(CUt_HashTable *tablePtr);
*/ 
#define CU_GetHashEntryCount(tablePtr)  ((tablePtr)->numEntries)

/**************************************************************************/
/* OTHER MACROS */

/* Macros to use for clients to use to invoke find and create procedures
   for hash tables: */

/* PROTO:
* CUt_HashEntry *CU_FindHashEntry( const CUt_HashTable *tablePtr, char *key ); 
*/
#define CU_FindHashEntry(tablePtr, key) \
	(*((tablePtr)->findProc))(tablePtr, key)

/* PROTO:
* CUt_HashEntry *CU_CreateHashEntry( CUt_HashTable *tablePtr, char *key, 
*                                    tInt4 *newPtr );
*/ 
#define CU_CreateHashEntry(tablePtr, key, newPtr) \
	(*((tablePtr)->createProc))(tablePtr, key, newPtr)


/**************************************************************************/
/* GLOBALS */


/**************************************************************************/
/* PROTOTYPES */

EXPORT_PREFIX void CU_InitHashTable( CUt_HashTable *tablePtr, 
                                     tUInt4 keyType );
EXPORT_PREFIX void CU_DeleteHashTable( CUt_HashTable *tablePtr );
EXPORT_PREFIX void CU_DeleteHashEntry( CUt_HashEntry *entryPtr );
EXPORT_PREFIX void CU_DeallocHashTable( CUt_HashTable *tablePtr );
EXPORT_PREFIX void CU_DeallocHashEntry( CUt_HashEntry *entryPtr );
EXPORT_PREFIX CUt_HashEntry * CU_FirstHashEntry( CUt_HashTable *tablePtr,
                                                 CUt_HashSearch *searchPtr );
EXPORT_PREFIX CUt_HashEntry * CU_NextHashEntry( CUt_HashSearch *searchPtr );
EXPORT_PREFIX char * CU_HashStats( CUt_HashTable *tablePtr );


#ifdef __cplusplus
} /* extern "C" */
#endif 

#endif /* CU_HASH_H_ALREADY_INCLUDED */
