/***************************************************************************
*
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*
* ==========================================================================
*
* FILE: CU_Stream.hh
*
* AUTHOR:
*
* YongJoon Lee
* Jaime Requena
* John Ehresman
* Stephan Deibel
*
* CREATION DATE:
*
* Aug 1995
*
* DESCRIPTION:
*
* Simple streams implementation, with slightly better error checking.
*
* NOTES:
*
*
* MODIFICATIONS:
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
*
* $Log: CU_Stream.hh,v $
* Revision 1.7  1998/12/01 23:18:50  bsloan
* Port to MetroWerks CodeWarrior Pro4 for Windows
*
* Revision 1.6  1998/11/20 20:42:24  swang
* *** empty log message ***
*
* Revision 1.5  1998/08/05 20:33:02  jehresma
* Implemented simple file output stream class
*
* Revision 1.4  1998/07/15 05:20:16  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
* Revision 1.3  1998/06/30 18:26:09  sdeibel
* Misc fixes to support compilation under gcc 2.7.2
*
// Revision 1.2  1997/12/04  17:06:08  jehresma
// Minor changes to suppress warnings
//
// Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
// Converting to CVS
//
* Revision 1.8  1997/09/15 21:58:35  jehresma
* Added virtual operators to OStream
*
* Revision 1.7  1997/07/01 15:11:25  sdeibel
* Added char text output operator / removed tInt1 numerical output
* operator (typecast to tInt2)
*
* Revision 1.6  1996/12/06  19:37:24  jehresma
* Added write method to CUC_OStrStream
*
* Revision 1.5  1996/09/23 23:03:21  sdeibel
* Misc
*
* Revision 1.4  1996/09/23  17:16:32  sdeibel
* Added OStream and OStrStream (written by Jaime; merged by me)
*
* Revision 1.3  1996/04/10  21:38:31  sdeibel
* Added better error checking;  Errors were being ignored and
* not passed out!
*
*
*/

#ifndef CU_STREAM_HH_ALREADY_INCLUDED
#define CU_STREAM_HH_ALREADY_INCLUDED


/**************************************************************************/
/* INCLUDES */

#include "CU_Basics.h"
#include "CU_Failure.h"
#include <stdio.h>
#include "CU_Text.h"


/**************************************************************************/
/* CLASSES */


class CUC_IStream {

public:

  CUC_IStream();
  virtual ~CUC_IStream();

  virtual tInt2 get();
  virtual tUInt4 read(void *dest, tUInt4 n);

  virtual tError seekg(tUInt4 pos);
  virtual tUInt4 tellg();

protected:

  const tInt1 *fBuff;
  const tInt1 *fBase; // start
  const tInt1 *fCPtr; // current
  const tInt1 *fEPtr; // end

};


class CUC_IFStream : public CUC_IStream {

public:

  CUC_IFStream();
  ~CUC_IFStream();

  tError open( const char *name );
  void close();

  tInt2 get();
  tUInt4 read(void *dest,tUInt4 n);

  tError seekg(tUInt4 pos);
  tUInt4 tellg();

private:

  char *fName;
  FILE *fFile;

};


class CUC_IStrStream : public CUC_IStream {

public:

  CUC_IStrStream(const char *buff, tUInt4 len);
  ~CUC_IStrStream();

};


class CUC_OStream {

public:

  CUC_OStream();
  virtual ~CUC_OStream();

  virtual CUC_OStream& operator << (char item) = 0;
  virtual CUC_OStream& operator << (char* item) = 0;
  virtual CUC_OStream& operator << (const char* item) = 0;
  virtual CUC_OStream& operator << (tInt2 item) = 0;
  virtual CUC_OStream& operator << (tInt4 item) = 0;
  virtual CUC_OStream& operator << (tUInt1 item) = 0;
  virtual CUC_OStream& operator << (tUInt2 item) = 0;
  virtual CUC_OStream& operator << (tUInt4 item) = 0;
  virtual CUC_OStream& operator << (tFloat item) = 0;

  tError GetError();

protected:

  tError fError;
};


class CUC_OStrStream : public virtual CUC_OStream {

public:

  CUC_OStrStream();
  virtual ~CUC_OStrStream();

  CUt_Text* fText;   /* the text */
  const char* str() const;

  CUC_OStrStream& write(const char* item, tUInt4 length);

  CUC_OStream& operator << (char item);
  CUC_OStream& operator << (char* item);
  CUC_OStream& operator << (const char* item);
  CUC_OStream& operator << (tInt2 item);
  CUC_OStream& operator << (tInt4 item);
  CUC_OStream& operator << (tUInt1 item);
  CUC_OStream& operator << (tUInt2 item);
  CUC_OStream& operator << (tUInt4 item);
  CUC_OStream& operator << (tFloat item);

};

class CUC_OFileStream : public virtual CUC_OStream {

public:

  CUC_OFileStream();
  CUC_OFileStream(FILE* fg);
  virtual ~CUC_OFileStream();

  CUC_OStream& write(const char* item, tUInt4 length);

  CUC_OStream& operator << (char item);
  CUC_OStream& operator << (char* item);
  CUC_OStream& operator << (const char* item);
  CUC_OStream& operator << (tInt2 item);
  CUC_OStream& operator << (tInt4 item);
  CUC_OStream& operator << (tUInt1 item);
  CUC_OStream& operator << (tUInt2 item);
  CUC_OStream& operator << (tUInt4 item);
  CUC_OStream& operator << (tFloat item);

private:
  FILE* fFP;
  tBoolean fCloseFP;
};


#endif /* CU_STREAM_HH_ALREADY_INCLUDED */
