/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Thread.prv.h
* 
* AUTHOR: 
*
* Stephan R.A. Deibel
* 
* CREATION DATE: 
* 
* 12/15/93
*
* VERSION: 
* 
* 8/4/94
*
* DESCRIPTION: 
* 
* A cross-platform threads package for multi-stack non-preemptive thread
* multitasking on a variety of hardware and compiler platforms.
* 
* NOTES: 
* 
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 12/15/93 Stephan   Created file / initial implementation
* 8/4/94   Stephan   Rewrote based on some code from Michael Hecht
*                    and some creative hacking to swap out stacks
*                    rather than copying them on each context switch.
*                    Also applied DSG naming and commenting conventions.
* 8/8/94   Stephan   Ported to SunOS 4.x.
* 8/10/94  Stepham   Augmented blocking facility.
* 9/21/94  Stephan   Added support for smart process yielding on non-
*                    preemptive systems.  Also some formatting/cosmetics.
* 7/18/95  Stephan   Ported to QuickThreads, so it will run on HP/UX 9.0.5
*                    among other systems.
* 
*/

#ifndef CU_THREAD_PRV_H_ALREADY_INCLUDED
#define CU_THREAD_PRV_H_ALREADY_INCLUDED

#ifndef CUk_NOTHREADS

#include <stddef.h>
/* #include <limits.h> -- Not needed? */

#ifdef macintosh
#include <setjmp.h>
#include <Events.h> 
#include <Memory.h>
#include <OSUtils.h>
#include <LowMem.h>
#endif
#ifdef MS_WIN
#include <windows.h>
#endif 
#if defined(SUNLWP)
#include <lwp/lwp.h>
#include <lwp/check.h>
#include <lwp/lwpmachdep.h>
#include <lwp/stackdep.h>
#endif
#ifdef QUICKTHREADS
#include <qt.h>
#endif
#ifdef MACTHREADMANAGER
#include <Threads.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/**************************************************************************/
/* CONSTANTS */

/* 
* Setjmp jump indicator
*/

#define CUk_RunThread 1

/*
* Process yielder recompute delay for non-preemptive multitasking systems
*/

#ifndef PREEMPTIVEMULTI
#define CUk_ProcYieldRedoDelay 30;
#endif

/*
* Stack sizes for QuickThreads implementation -- need to define them
* for each platform since QT_STKALIGN will vary on each machine and
* we probably want to use min size around 8K and default size around
* 32K.
*/

#ifdef QUICKTHREADS
#if defined(hpux9)
#define CUk_QThreadMinStack QT_STKALIGN * 256
#define CUk_QThreadDefaultStack QT_STKALIGN * 512
#elif defined(linux)
#define CUk_QThreadMinStack QT_STKALIGN * 256
#define CUk_QThreadDefaultStack QT_STKALIGN * 512
#elif defined(sunos4)
#define CUk_QThreadMinStack QT_STKALIGN * 256
#define CUk_QThreadDefaultStack QT_STKALIGN * 512
#endif
#endif

/*
* Stack size definitions for win32.  Sizes are in bytes.
*/

#ifdef WIN32
#define CUk_Win32ThreadMinStack     8000
#define CUk_Win32ThreadDefaultStack 32000
#endif


/**************************************************************************/
/* TYPES */

/* 
* Register ordering (see jmpbuf.c).  Actually, followed by FP regs,
* which aren't included here since we don't access them anyway. 
*
* NOTE: These are of course compiler / system dependent and are only
* used in the SETJMP implementation of CU_Thread.
*
*/

#ifdef SETJMPTHREADS
#ifdef macintosh

#if defined(THINK_C)
enum {
  d3, d4, d5, d6, d7,
  a1, a2, a3, a4, a6, a7
};
#elif defined(__MWERKS__)
enum {
  d2 = 0, d3 = 1, d4 = 2, d5 = 3, d6 = 4, d7 = 5,
  pc = 6, a2 = 7, a3 = 8, a4 = 9, a6 = 10, a7 = 11
};
#endif /* THINK_C or MWERKS */
#endif /* macintosh */

#ifdef MS_WIN
#ifdef MS_VC
enum {
  sp = 3, pc = 4, ds = 6
};
#endif /* MS_WIN */
#endif /* MS_VC */
#endif /* SETJMPTHREADS */

/*
* Thread status
*/

typedef enum {
    CUk_Ready = 0L,       /* Ready to execute */
    CUk_Running = 1L,     /* Currently executing */
    CUk_Blocked = 2L,     /* Waiting for resources or sleeping */
    CUk_Destroyed = 3L    /* Thread has been destroyed */
} CUt_ThreadStatus;

/*
* Thread environment
*/

typedef struct CUt_ThreadEnv {
  void *stack;
#ifdef SETJMPTHREADS
  jmp_buf setjmp_env;
#endif
#if defined(macintosh) && defined(SETJMPTHREADS)
  void *heap_end;
  void *appl_limit;
  void *hi_heap_mark;
#endif
#ifdef MS_WIN
#ifdef DEBUG 
  int env0;
  int pc0;
  int pc1;
  int sp;   
#endif
#endif /* MS_WIN */
} CUt_ThreadEnv;

/*
* Thread list entry
*/

typedef struct CUt_ThreadEntry {
  struct CUt_ThreadEntry *next_entry;
  struct CUt_ThreadEntry *prev_entry;
  
  CUt_ThreadID ID;
  CUt_ThreadStatus status;
  CUt_ThreadProc proc;
  CUt_ThreadEnv env;
  void *user_data;
  
  /* Blocking data; some or all ignored based on status and block_opt */
  CUt_BlockOpt block_opt;
  tUInt4 wake_time;
  tBoolean *blocked_bool;
  CUt_ThreadBlockProc blocked_fct;
  void *blocked_params;
  tUInt4 check_freq;
  void *blocked_mem;
  void *target_mem;
  tUInt4 mem_len;
  tBoolean local_peer;

/* PROBABLY NOT NEEDED! */
#if 0
  /* Run statistics; used to predict future load */
  tUInt4 last_run;
#endif

#if defined(SUNLWP)
  thread_t lwp_ID;
#endif
#ifdef QUICKTHREADS
  qt_t *qthread;
#endif 
#if defined(MACTHREADMANAGER)
  ThreadID tm_ID;
#endif
#ifdef WIN32THREADS
  HANDLE win32_thread;
  HANDLE win32_event;
  tBoolean suspended;
#endif
} CUt_ThreadEntry;

/*
* Thread list control structure
*/

typedef struct CUt_ThreadControl {
  CUt_ThreadID ID_seed;
  tUInt4 num_local_peers;
  
  CUt_ThreadEntry *master;
  CUt_ThreadEntry *current;
  
  CUt_ThreadEntry *destroy;
} CUt_ThreadControl;

/* 
* Mac Thread Manager specific type for translation between CUt_ThreadProc
* formatted proc and the proc type used by thread manager
*/

#if defined(MACTHREADMANAGER)
typedef struct CUt_MacThreadProcData {
  CUt_ThreadProc proc;
  void *data;
  CUt_ThreadEntry *thread;
} CUt_MacThreadProcData;
#endif


/**************************************************************************/
/* ACCESS MACROS */

/*
* Macintosh only:
*
* Macros for faster access to macintosh LM globals (mostly defined in 
* LoMem.h but these are missing from there).
*/

#if defined(macintosh) && defined(SETJMPTHREADS)
Ptr HiHeapMark : 0x0BAE;
Ptr StkLowPt : 0x0110;
#define LMSetHiHeapMark(p)    ((void) (HiHeapMark = (p)))
#define LMGetHiHeapMark()     (HiHeapMark)
#define LMSetStkLowPt(p)      ((void) (StkLowPt = (p)))
#endif


/**************************************************************************/
/* OTHER MACROS */

/* Dummy macro for stripping code on preemptive multitasking systems */
#ifdef PREEMPTIVEMULTI
#define CU_WakeUpProcessYielder() ;
#endif

/* Stack alignment macro for QuickThreads implementation */
#ifdef QUICKTHREADS
#define CU_STKALIGN(sp) \
  ((void *)((((qt_word_t)(sp)) + (QT_STKALIGN) - 1) & ~((QT_STKALIGN)-1)))
#endif


/**************************************************************************/
/* GLOBALS */

/*
* The thread control structure - contains information about the state of
* the thread package, including the table of all threads.
*/

extern CUt_ThreadControl CUg_ThreadControl;

/*
* Process dependency list - contains information about external processes
* engaged in communication with local threads and information about which
* threads are blocked waiting for an external process.  Used to do smart
* time slicing to other local processes on cooperative multitasking
* systems.
*/

#ifndef PREEMPTIVEMULTI
extern tUInt4 CUg_NextProcYield;
extern tUInt4 CUg_ProcYieldFreq;
extern tUInt4 CUg_ProcYieldDuration;
extern tBoolean CUg_ProcYieldStatusChanged;
extern tUInt4 CUg_ProcYieldRedoTime;
extern CUt_ThreadEntry *CUg_ProcYielderThread;
#endif


/**************************************************************************/
/* PROTOTYPES */

/*
* Thread list management utilities
*/

void CU_InsertTLEntry(CUt_ThreadEntry *thread);
void CU_RemoveTLEntry(CUt_ThreadEntry *thread);
CUt_ThreadEntry *CU_FindTLEntry(register CUt_ThreadID ID);

/*
* Thread creation / deletion
*/

void CU_DestroyThreadEntry(CUt_ThreadEntry *thread);

/*
* Misc local utilities
*/

size_t CU_MinStackSize();
size_t CU_DefaultStackSize();

/*
* Process yielder utilities - non-preemptive multitasking systems only
*/

#ifndef PREEMPTIVEMULTI
void CU_WakeUpProcessYielder();
#endif

/*
* Thread control
*/

void CU_SaveEnvironment();
void CU_RestoreEnvironment();

void CU_BlockThreadEntry(CUt_ThreadEntry *thread, CUt_BlockOpt opt, void *p1, 
  void *p2, void *p3, void *p4, tBoolean local_peer);

tBoolean CU_ThreadEntryReady(CUt_ThreadEntry *thread, tUInt4 ticks);
CUt_ThreadEntry *CU_NextScheduledThreadEntry();
tBoolean CU_StartThreadEntry(CUt_ThreadEntry *thread);

void EXPORT_PREFIX CU_ThreadYield();

/*
* Sun OS-dependent utilities
*/

#if defined(SUNLWP)
void CU_SunThreadProc(CUt_ThreadProc proc, void *data, CUt_ThreadEntry *thread);
#endif

/*
* Mac Thread Manager-dependent utilities
*/

#if defined(MACTHREADMANAGER)
pascal void *CU_MacThreadProc(CUt_MacThreadProcData *data);
#endif

#if defined(WIN32THREADS)
void
CU_StartWin32Thread(CUt_ThreadEntry* start_thread);
void
CU_DestroyWin32Thread(CUt_ThreadEntry* thread);
#endif

#ifdef __cplusplus
} /* extern "C" */
#endif 

#endif /* CUk_NOTHREADS */

#endif /* CU_THREAD_PRV_H_ALREADY_INCLUDED */
